import gomea
import numpy as np
from functools import cache

# Custom fitness function resembling the concatenated deceptive trap function of size k
class RosenbrockFunctionGBO(gomea.fitness.GBOFitnessFunctionRealValued):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    @cache
    def number_of_subfunctions( self ):
        return self.number_of_variables - 1
    
    @cache
    def inputs_to_subfunction( self, subfunction_index ):
        return [subfunction_index, subfunction_index+1]

    def subfunction(self, subfunction_index, variables):
        x = variables[subfunction_index]
        y = variables[subfunction_index+1]
        return( 100*(y-x*x)*(y-x*x) + (1.0-x)*(1.0-x) )

class RosenbrockFunctionBBO(gomea.fitness.BBOFitnessFunctionRealValued):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def objective_function(self, objective_index, variables):
        result = 0.0
        for i in range(self.number_of_variables-1):
            x = variables[i]
            y = variables[i+1]
            result += 100*(y-x*x)*(y-x*x) + (1.0-x)*(1.0-x)
        return result
    
    def constraint_function(self, variables):
        return 0.0


import gomea
import numpy as np

# Custom fitness function resembling the concatenated deceptive trap function of size k
class CustomTrapFunction(gomea.fitness.BBOFitnessFunctionDiscrete):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, k, **kwargs):
        assert( number_of_variables % k == 0 )
        self.k = k
        return super().__new__(self,number_of_variables,**kwargs)

    def objective_function(self, objective_index, variables):
        f = 0
        for i in range(0,self.number_of_variables,self.k):
            trap_variables = variables[i:i+self.k]
            unitation = np.sum(trap_variables)
            if unitation == self.k:
                f += unitation
            else:
                f += self.k - unitation - 1
        return f

dim = 20
lm = gomea.linkage.BlockMarginalProduct(block_size=5)
fd = CustomTrapFunction(dim,k=5,value_to_reach=dim)
dgom = gomea.DiscreteGOMEA(fitness=fd,linkage_model=lm,max_number_of_evaluations=100000)
result = dgom.run()
result.printAllStatistics()



import gomea.fitness
import gomea.real_valued
import gomea.discrete
import gomea.linkage

def RealValuedGOMEA(**kwargs):
    return gomea.real_valued.RealValuedGOMEA(**kwargs)

def DiscreteGOMEA(**kwargs):
    return gomea.discrete.DiscreteGOMEA(**kwargs)


import json

def getResDir(prob_tag):
    res_dir = "results/"
    if prob_tag == "trap5":
        res_dir = "deceptive_trap/" + res_dir
    elif prob_tag == "sorebchainstrong":
        res_dir = "soreb_chain/" + res_dir
    else:
        res_dir = prob_tag + "/" + res_dir
    return res_dir

def writeResultsToJson(input,prob_tag,lang_tag,lm_tag):
    res_dir = getResDir(prob_tag)
    full_tag = prob_tag+"_"+lang_tag+"_"+lm_tag
    filename = res_dir+full_tag+".json"
    with open(filename,'w') as fp:
        json.dump(input,fp)

def readResultsFromJson(prob_tag,lang_tag,lm_tag):
    res_dir = getResDir(prob_tag)
    full_tag = prob_tag+"_"+lang_tag+"_"+lm_tag
    filename = res_dir+full_tag+".json"
    try:
        with open(filename,'r') as fp:
            res = json.load(fp)
            for k,sub_dict in res.items():
                try:
                    res[k] = {int(dim):[float(r) for r in res] for dim,res in sub_dict.items()}
                except Exception as e:
                    res[k] = {int(dim):res for dim,res in sub_dict.items()}
            return res
    except Exception as e:
        print(e)
        return {}

def resultsExist(prob_tag,lang_tag,lm_tag):
    res_dir = getResDir(prob_tag)
    full_tag = prob_tag+"_"+lang_tag+"_"+lm_tag
    filename = res_dir+full_tag+".json"
    try:
        with open(filename,'r') as fp:
            return True
    except FileNotFoundError:
        return False


# distutils: language = c++
# cython: c_string_type=unicode, c_string_encoding=utf8

from gomea.real_valued cimport rvg_t, Config
from gomea.output cimport OutputStatisticsWrapper
from gomea.output import OutputStatistics
from libcpp.string cimport string
from libcpp cimport bool
import inspect

from gomea.fitness cimport FitnessFunction, GBOFitnessFunction, BBOFitnessFunction
from gomea.linkage cimport LinkageModel, StaticLinkageTree
include "EmbeddedFitness.pxi"

# Create a Cython extension type which holds a C++ instance
# as an attribute and create a bunch of forwarding methods
# Python extension type.
cdef class RealValuedGOMEA:
    cdef rvg_t c_inst  # Hold a C++ instance which we're wrapping
    cdef Config c_config
    cdef FitnessFunction fitness

    def __cinit__(self,
        # Optimization problem settings (required)
        fitness: FitnessFunction, 
        # GOMEA settings (optional)
        linkage_model : LinkageModel = StaticLinkageTree(),
        lower_init_range: float = 0.0,
        upper_init_range: float = 1.0,
        random_seed: int=-1,
        # IMS settings (optional)
        max_number_of_populations: int = 25,
        IMS_subgeneration_factor: int = 8,
        base_population_size: int = 10,
        # Termination settings (optional)
        max_number_of_generations: int = -1,
        max_number_of_evaluations : float = -1,
        max_number_of_seconds : float = -1,
        # Output settings (optional)
        generational_statistics = True,
        generational_solution = False,
        #output_frequency = 'GEN',
        verbose = False,
    ):
        
        # Initialize attributes 
        self.c_config = Config()
        self.c_config.problem_index = 0
        assert( (<FitnessFunction?>fitness).c_inst_realvalued != NULL, "FitnessFunction is not real-valued." )
        self.c_config.fitness = (<FitnessFunction?>fitness).c_inst_realvalued
        # as we only store a pointer to a field of the FitnessFunction (which actually manages the memory around the pointer)
        # there is the risk of the FitnessFunction being removed & its contents deallocated. This leaves the above pointer dangling.
        # This is particularly likely if the FitnessFunction was constructed as an argument to this class -
        # in which case it & corresponding memory will be deallocated after this initialization procedure has completed.
        # Fix this by keeping the reference to the FitnessFunction around for as long as this instance exists.
        self.fitness = fitness

        self.c_config.linkage_config = linkage_model.c_inst
        self.c_config.use_vtr = True 
        self.c_config.lower_user_range = lower_init_range
        self.c_config.upper_user_range = upper_init_range
        self.c_config.tau = 0.35
        self.c_config.distribution_multiplier_decrease = 0.9
        self.c_config.st_dev_ratio_threshold = 1.0
        self.c_config.fitness_variance_tolerance = 0.0
        self.c_config.maximum_no_improvement_stretch = 100
        self.c_config.selection_during_gom = True
        self.c_config.update_elitist_during_gom = True 
        self.c_config.maximum_number_of_populations = max_number_of_populations
        self.c_config.number_of_subgenerations_per_population_factor = IMS_subgeneration_factor
        self.c_config.base_population_size = base_population_size
        self.c_config.maximum_number_of_generations = max_number_of_generations
        self.c_config.maximum_number_of_evaluations = max_number_of_evaluations
        self.c_config.maximum_number_of_seconds = max_number_of_seconds
        self.c_config.black_box_evaluations = False
        self.c_config.generational_statistics = generational_statistics 
        self.c_config.generational_solution = generational_solution
        self.c_config.verbose = verbose
        self.c_config.print_verbose_overview = False 
        output_frequency = 'GEN'
        if output_frequency == 'GEN':
            self.c_config.output_frequency = GEN
        elif output_frequency == 'IMS_GEN':
            self.c_config.output_frequency = IMS_GEN
        elif output_frequency == 'NEW_ELITE':
            self.c_config.output_frequency = NEW_ELITE
        else:
            raise ValueError("output_frequency must be one of 'GEN', 'IMS_GEN', 'NEW_ELITE'")
        self.c_config.fix_seed = 0
        if random_seed != -1:
            self.c_config.random_seed = random_seed
            self.c_config.fix_seed = 1



        # Initialize C++ instance
        self.c_inst = rvg_t(&self.c_config)
    
    def print_usage(self):
        self.c_inst.printUsage()

    def run(self):
        self.c_inst.run()
        return OutputStatistics(OutputStatisticsWrapper.from_ptr(&self.c_inst.output))



import gomea
import numpy as np
from functools import cache

# Custom fitness function resembling the concatenated deceptive trap function of size k
class SOREBChainStrongGBO(gomea.fitness.GBOFitnessFunctionRealValued):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, value_to_reach ):
        self.rotation_angle = 45
        self.rotation_block_size = 2
        return super().__new__(self,number_of_variables,value_to_reach)

    @cache
    def number_of_subfunctions( self ):
        return self.number_of_variables - 1
    
    @cache
    def inputs_to_subfunction( self, subfunction_index ):
        return [subfunction_index, subfunction_index+1]

    def subfunction(self, subfunction_index, variables):
        rotated_variables = self.rotate_variables(variables[self.inputs_to_subfunction(subfunction_index)],self.rotation_angle)
        result = 0.0
        for i in range(self.rotation_block_size):
            result += 10.0**(6.0*(i/(self.rotation_block_size-1) ))*rotated_variables[i]*rotated_variables[i]
        return result

class SOREBChainStrongBBO(gomea.fitness.BBOFitnessFunctionRealValued):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, value_to_reach ):
        self.rotation_angle = 45
        self.rotation_block_size = 2
        return super().__new__(self,number_of_variables,value_to_reach)

    def objective_function(self, objective_index, variables):
        result = 0.0
        for i in range(self.number_of_variables-1):
            rotated_variables = self.rotate_variables(variables[i:i+2],self.rotation_angle)
            for i in range(self.rotation_block_size):
                result += 10.0**(6.0*(i/(self.rotation_block_size-1) ))*rotated_variables[i]*rotated_variables[i]
        return result
    
    def constraint_function(self, variables):
        return 0.0


import gomea
import numpy as np
import gc

# Custom fitness function resembling the Rosenbrock function
class CustomRastriginFunction(gomea.fitness.GBOFitnessFunctionRealValued):
    def number_of_subfunctions( self ):
        return self.number_of_variables
    
    def inputs_to_subfunction( self, subfunction_index ):
        return [subfunction_index]

    def objective_function( self, objective_index, fitness_buffers ):
        return 10.0 * self.number_of_variables + fitness_buffers[0]

    def subfunction(self, subfunction_index, variables):
        x = variables[subfunction_index]
        return x*x - 10.0*np.cos(2.0*np.pi*x)

# Custom fitness function resembling the concatenated deceptive trap function of size k
class CustomTrapFunction(gomea.fitness.GBOFitnessFunctionDiscrete):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, k, value_to_reach):
        assert( number_of_variables % k == 0 )
        self.k = k
        return super().__new__(self,number_of_variables,value_to_reach)

    def number_of_subfunctions( self ):
        return self.number_of_variables // self.k
    
    def inputs_to_subfunction( self, subfunction_index ):
        return range(self.k*subfunction_index,self.k*subfunction_index+self.k)

    def subfunction(self, subfunction_index, variables):
        trap_variables = variables[self.inputs_to_subfunction(subfunction_index)]
        unitation = np.sum(trap_variables)
        if unitation == self.k:
            return unitation
        else:
            return self.k - unitation - 1

dim = 20
vtr = 1e-10
#frv = gomea.fitness.SphereFunction(10000,value_to_reach=vtr)
#frv = gomea.fitness.RosenbrockFunction(20,value_to_reach=vtr)
frv = CustomRastriginFunction(dim,value_to_reach=vtr)

lm = gomea.linkage.Univariate()
nsucc = 0
nruns = 5
rvgom = gomea.RealValuedGOMEA(fitness=frv,linkage_model=lm,lower_init_range=5,upper_init_range=10, max_number_of_populations=1, base_population_size=200, max_number_of_evaluations=100000)
print("ObjVal\tNumEvaluations\tTime(s)")
for i in range(nruns):
    result = rvgom.run()
    print(result['best_obj_val'][-1],result['evaluations'][-1],result['time'][-1])
    if result['best_obj_val'][-1] < vtr:
        nsucc += 1
print(nsucc,"/",nruns," successes")
#result.printFinalStatistics()
result.printAllStatistics()

#fd = gomea.fitness.OneMaxFunction(1000)
#lm = gomea.linkage.StaticLinkageTree(maximum_set_size=5)
#lm = gomea.linkage.Custom(file="FOS.in")
#lm = gomea.linkage.Custom(fos=[range(0,5),range(5,10),range(10,15),range(15,20)])
#lm = gomea.linkage.BlockMarginalProduct(block_size=5)
lm = gomea.linkage.LinkageTree()
#fd = CustomTrapFunction(20,k=5,value_to_reach=20)
dim = 320
fd = gomea.fitness.DeceptiveTrapFunction(dim,trap_size=5)
#fd = CustomTrapFunction(dim,k=5,value_to_reach=dim)
nsucc = 0
print("ObjVal\tNumEvaluations\tTime(s)")
for i in range(nruns):
    dgom = gomea.DiscreteGOMEA(fitness=fd,linkage_model=lm,max_number_of_evaluations=1000000)
    result = dgom.run()
    #result.printFinalStatistics()
    print(result['best_obj_val'][-1],result['evaluations'][-1],result['time'][-1])
    if result['best_obj_val'][-1] == dim:
        nsucc += 1
print(nsucc,"/",nruns," successes")
result.printAllStatistics()



import gomea
import numpy as np
import cython
from functools import cache

# Custom fitness function resembling the concatenated deceptive trap function of size k
class DeceptiveTrapFunction(gomea.fitness.GBOFitnessFunctionDiscrete):
    k : cython.int
    
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, k ):
        assert( number_of_variables % k == 0 )
        self.k = k
        value_to_reach = number_of_variables
        return super().__new__(self,number_of_variables,value_to_reach)

    @cache
    def number_of_subfunctions( self ) -> cython.int:
        return self.number_of_variables // self.k
    
    @cache
    def inputs_to_subfunction( self, subfunction_index : cython.int ) -> np.ndarray:
        return range(self.k*subfunction_index,self.k*subfunction_index+self.k)

    def subfunction(self, subfunction_index : cython.int, variables : np.ndarray ):
        trap_variables : np.ndarray = variables[self.inputs_to_subfunction(subfunction_index)]
        unitation : cython.int = np.sum(trap_variables)
        if unitation == self.k:
            return unitation
        else:
            return self.k - unitation - 1


import gomea
import matplotlib.pyplot as plt

frv = gomea.fitness.RosenbrockFunction(10,value_to_reach=1e-10)
lm = gomea.linkage.Univariate()
rvgom = gomea.RealValuedGOMEA(fitness=frv, linkage_model=lm, lower_init_range=-115, upper_init_range=-100)
result = rvgom.run()
plt.grid()
plt.yscale('log')
plt.xscale('log')
plt.xlabel('Number of evaluations')
plt.ylabel('Best objective value')
plt.plot(result['evaluations'],result['best_obj_val'])
plt.savefig('convergence_plot.pdf')




import gomea
import numpy as np
from functools import cache

# Custom fitness function resembling the concatenated deceptive trap function of size k
class DeceptiveTrapFunctionGBO(gomea.fitness.GBOFitnessFunctionDiscrete):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, k ):
        assert( number_of_variables % k == 0 )
        self.k = k
        value_to_reach = number_of_variables
        return super().__new__(self,number_of_variables,value_to_reach)

    @cache
    def number_of_subfunctions( self ):
        return self.number_of_variables // self.k
    
    @cache
    def inputs_to_subfunction( self, subfunction_index ):
        return range(self.k*subfunction_index,self.k*subfunction_index+self.k)

    def subfunction(self, subfunction_index, variables):
        trap_variables = variables[self.inputs_to_subfunction(subfunction_index)]
        unitation = np.sum(trap_variables)
        if unitation == self.k:
            return unitation
        else:
            return self.k - unitation - 1

class DeceptiveTrapFunctionBBO(gomea.fitness.BBOFitnessFunctionDiscrete):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, k ):
        assert( number_of_variables % k == 0 )
        self.k = k
        value_to_reach = number_of_variables
        return super().__new__(self,number_of_variables,value_to_reach)

    def objective_function(self, objective_index, variables):
        def trap_function(unitation):
            if unitation == self.k:
                return unitation
            else:
                return self.k - unitation - 1
        
        start_indices = np.arange(0,self.number_of_variables,self.k)
        unitations = [np.sum(variables[ind:ind+self.k]) for ind in start_indices]
        f = np.sum([trap_function(u) for u in unitations])
        return f
    
    def constraint_function(self, variables):
        return 0.0


from experiments import runExperiment
from output import readResultsFromJson
from plotting import plotEvalTimeScalability, plotTimePerFEV, plotTimeScalability, plotFEVScalability

def runExperimentsMaxCutBFLT(bound,max_dim=-1,rerun=False,nthreads=1,nruns=10,max_sec=-1,input_file="",vtr_file=""):
    runExperiment('maxcut','cy','lt'+str(bound),max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    #runExperiment('maxcut','cyc','lt'+str(bound),max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','cpp','lt'+str(bound),max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','cpp_bbo','lt'+str(bound),max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','py','lt'+str(bound),max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','py_bbo','lt'+str(bound),max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)

def runExperimentsMaxCutLT(max_dim=-1,rerun=False,nthreads=1,nruns=10,max_sec=-1,input_file="",vtr_file=""):
    runExperiment('maxcut','cy','lt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    #runExperiment('maxcut','cyc','lt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','cpp','lt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','cpp_bbo','lt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','py','lt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','py_bbo','lt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)

def runExperimentsMaxCutStaticLT(max_dim=-1,rerun=False,nthreads=1,nruns=10,max_sec=-1,input_file="",vtr_file=""):
    runExperiment('maxcut','py','slt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','cy','slt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    #runExperiment('maxcut','cyc','slt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)
    runExperiment('maxcut','cpp','slt',max_dim=-1,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec,input_file=input_file,vtr_file=vtr_file)

def runExperiments(input_dir):
    rerun_experiments = False 
    nthreads = 15
    nruns = 30
    max_sec = 600
    dims = [9,16,25,49,100,196,400,784,1600]
    for d in dims:
        input_file = f'{input_dir}n{d:07d}i00.txt'
        vtr_file = f'{input_dir}n{d:07d}i00.bkv'
        runExperimentsMaxCutLT(max_dim=-1,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,input_file=input_file,vtr_file=vtr_file)
        runExperimentsMaxCutStaticLT(max_dim=-1,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,input_file=input_file,vtr_file=vtr_file)
        #runExperimentsMaxCutBFLT(10,max_dim=-1,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,input_file=input_file,vtr_file=vtr_file)
        runExperimentsMaxCutBFLT(50,max_dim=-1,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,input_file=input_file,vtr_file=vtr_file)
        runExperimentsMaxCutBFLT(100,max_dim=-1,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,input_file=input_file,vtr_file=vtr_file)

def makePlots(out_dir="plots/"):
    max_dim = 2500
    res_py_lt = readResultsFromJson('maxcut','py','lt')
    #res_cy_lt = readResultsFromJson('maxcut','cy','lt')
    #res_cyc = readResultsFromJson('maxcut_cyc_lt')
    res_cpp_lt = readResultsFromJson('maxcut','cpp','lt')
    res_pybbo_lt = readResultsFromJson('maxcut','py_bbo','lt')
    res_cppbbo_lt = readResultsFromJson('maxcut','cpp_bbo','lt')
    results = [("GBO-Python",res_py_lt),("GBO-C++",res_cpp_lt),("BBO-Python",res_pybbo_lt),("BBO-C++",res_cppbbo_lt)]
    #plotTimePerFEV(results,out_dir=out_dir,tag='maxcut_lt',max_dim=max_dim)
    #plotTimeScalability(results,out_dir=out_dir,tag='maxcut_lt',max_dim=max_dim)
    plotEvalTimeScalability(results,out_dir=out_dir,tag='maxcut_lt',max_dim=max_dim)
    #plotFEVScalability(results,out_dir=out_dir,tag='maxcut_lt',max_dim=max_dim)
    
    res_py_slt = readResultsFromJson('maxcut','py','slt')
    res_cpp_slt = readResultsFromJson('maxcut','cpp','slt')
    #results = [("GBO-Python",res_py_slt),("GBO-C++",res_cpp_slt),("BBO-Python",res_pybbo_lt),("BBO-C++",res_cppbbo_lt)]
    #plotTimePerFEV(results,out_dir=out_dir,tag='maxcut_slt',max_dim=max_dim)
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='maxcut_slt',max_dim=max_dim)
    #plotTimeScalability(results,out_dir=out_dir,tag='maxcut_slt',max_dim=max_dim)
    #plotFEVScalability(results,out_dir=out_dir,tag='maxcut_slt',max_dim=max_dim)
    
    #res_py_lt50 = readResultsFromJson('maxcut','py','lt50')
    #res_py_lt100 = readResultsFromJson('maxcut','py','lt100')
    res_pybbo_lt = readResultsFromJson('maxcut','py_bbo','lt')
    res_pybbo_lt50 = readResultsFromJson('maxcut','py_bbo','lt50')
    #res_pybbo_lt100 = readResultsFromJson('maxcut','py_bbo','lt100')
    results = [("GBO-LT",res_py_lt),("GBO-SLT",res_py_slt),("BBO-LT",res_pybbo_lt),("BBO-LT(50)",res_pybbo_lt50)]
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='maxcut_py_lms',max_dim=max_dim)
    plotTimeScalability(results,out_dir=out_dir,tag='maxcut_py_lms',max_dim=max_dim)
    #plotFEVScalability(results,out_dir=out_dir,tag='maxcut_py_lms',max_dim=max_dim)
    
    res_cpp_lt50 = readResultsFromJson('maxcut','cpp','lt50')
    res_cpp_lt100 = readResultsFromJson('maxcut','cpp','lt100')
    res_cppbbo_lt = readResultsFromJson('maxcut','cpp_bbo','lt')
    res_cppbbo_lt50 = readResultsFromJson('maxcut','cpp_bbo','lt50')
    res_cppbbo_lt100 = readResultsFromJson('maxcut','cpp_bbo','lt100')
    #results = [("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50),("BBO-LT(100)",res_cppbbo_lt100),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt),("GBO-LT(50)",res_cpp_lt50),("GBO-LT(100)",res_cpp_lt100)]
    results = [("GBO-LT",res_cpp_lt),("GBO-SLT",res_cpp_slt),("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50)]
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='maxcut_cpp_lms',max_dim=max_dim)
    #plotTimeScalability(results,out_dir=out_dir,tag='maxcut_cpp_lms',max_dim=max_dim)
    plotFEVScalability(results,out_dir=out_dir,tag='maxcut_cpp_lms',max_dim=max_dim)

if __name__ == '__main__':
    #runExperiments("problem_data/maxcut/set0c/")
    makePlots("plots/set0c/")


from gomea.output cimport output_statistics_t
from ctypes import * 

cdef class OutputStatisticsWrapper:
    def __cinit__(self):
        pass

    def __init__(self):
        raise TypeError("This class cannot be instantiated directly.")

    @staticmethod
    cdef OutputStatisticsWrapper from_ptr(output_statistics_t *_ptr):
        cdef OutputStatisticsWrapper wrapper = OutputStatisticsWrapper.__new__(OutputStatisticsWrapper)
        wrapper.c_ptr = _ptr
        return wrapper

class OutputStatistics:
    def __init__(self, stats : OutputStatisticsWrapper):
        self.generational_metrics = stats.c_ptr[0].getGenerationalMetricNames()
        self.generational_metrics_dict = {}
        self.final_metrics = stats.c_ptr[0].getFinalMetricNames()
        self.final_metrics_dict = {}

        def getGenerationalMetricValues( metric_name ):
            try:
                return stats.c_ptr[0].getGenerationalMetricValues[int](metric_name)
            except Exception as err:
                pass
            try:
                return stats.c_ptr[0].getGenerationalMetricValues[double](metric_name)
            except Exception as err:
                pass
            try:
                b_result = stats.c_ptr[0].getGenerationalMetricValues[string](metric_name)
                return [byte_string.decode('utf-8') for byte_string in b_result]
            except Exception as err:
                pass

        def getFinalMetricValue( metric_name ):
            try:
                return stats.c_ptr[0].getFinalMetricValue[int](metric_name)
            except Exception as err:
                pass
            try:
                return stats.c_ptr[0].getFinalMetricValue[double](metric_name)
            except Exception as err:
                pass
            try:
                b_result = stats.c_ptr[0].getFinalMetricValue[string](metric_name)
                return b_result.decode('utf-8')
            except Exception as err:
                pass

        for metric_name in self.generational_metrics:
            s_metric_name = metric_name.decode()
            self.generational_metrics_dict[s_metric_name] = getGenerationalMetricValues(metric_name)
        self.generational_metrics = [s.decode() for s in self.generational_metrics]
        for metric_name in self.final_metrics:
            s_metric_name = metric_name.decode()
            self.final_metrics_dict[s_metric_name] = getFinalMetricValue(metric_name)
        self.final_metrics = [s.decode() for s in self.final_metrics]

    def __getitem__(self, metric):
        assert( metric in self.generational_metrics, str(metric)+" is not a valid metric." )
        return self.generational_metrics_dict[metric]

    def getFinalStatistics(self):
        return self.final_metrics_dict.copy()

    def getGenerationalStatistics(self):
        return self.generational_metrics_dict.copy()
    
    def printFinalStatistics(self):
        print("Final statistics:")
        for k,v in self.final_metrics_dict.items():
            print(k,":",v)

    def printGenerationalStatistics(self):
        print("Generational statistics:")
        for k,v in self.generational_metrics_dict.items():
            print(k,":",v)
    
    def printAllStatistics(self):
        self.printGenerationalStatistics()
        self.printFinalStatistics()



from libcpp.map cimport map
from libcpp.string cimport string
from libcpp.vector cimport vector

cdef extern from "gomea/src/common/output_statistics.hpp" namespace "gomea":
    cdef cppclass output_statistics_t:
        output_statistics_t() except +
        vector[string] getGenerationalMetricNames() except +
        vector[T] getGenerationalMetricValues[T](string) except +
        vector[string] getFinalMetricNames() except +
        T getFinalMetricValue[T](string) except +

cdef class OutputStatisticsWrapper:
    cdef output_statistics_t *c_ptr
    
    @staticmethod
    cdef OutputStatisticsWrapper from_ptr(output_statistics_t *_ptr)



from cpython cimport PyObject
from libcpp.string cimport string
from libcpp.pair cimport pair

include "EmbeddedFitness.pxi"
    
cdef class FitnessFunction:
    def __cinit__(self, 
        *args,
        **kwargs
    ):
        c_inst_discrete = NULL
        c_inst_realvalued = NULL
    
    def __init__(self, 
        *args,
        **kwargs
    ):
        self.rotation_matrix = np.ndarray(0)

    cpdef void initialize_rotation_matrix(self, int rotation_block_size, double rotation_angle):
        rotation_matrix : np.ndarray = np.identity(rotation_block_size)
        theta : float = np.radians(rotation_angle)
        cos_theta : float = np.cos(theta)
        sin_theta : float = np.sin(theta)
        for index0 in range(rotation_block_size-1):
            for index1 in range(index0+1,rotation_block_size):
                matrix = np.identity(rotation_block_size)
                matrix[index0][index0] = cos_theta
                matrix[index0][index1] = -sin_theta
                matrix[index1][index0] = sin_theta
                matrix[index1][index1] = cos_theta
                rotation_matrix = np.matmul(matrix, rotation_matrix)
        self.rotation_matrices[(rotation_block_size,rotation_angle)] = <PyObject*> rotation_matrix
        self.rotation_matrix = rotation_matrix

    cpdef np.ndarray rotate_variables(self, np.ndarray variables, double rotation_angle):
        key = (len(variables),rotation_angle)
        if(self.rotation_matrices.count(key) == 0):
            self.initialize_rotation_matrix(len(variables), rotation_angle)
        rotation_matrix = <np.ndarray?> self.rotation_matrices[(len(variables),rotation_angle)]
        rotated_variables = np.matmul(rotation_matrix, variables)
        return rotated_variables
        

cdef class YourFitnessFunctionDiscrete(FitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        alphabet_size : int = 2,
        value_to_reach : float = 1e308
    ):
        self.number_of_variables = number_of_variables
        self.alphabet_size = alphabet_size
        self.value_to_reach = value_to_reach
        self.c_inst_discrete = new yourFitnessFunctionDiscrete(number_of_variables,alphabet_size,value_to_reach)

    def __dealloc__(self):
        del self.c_inst_discrete

cdef class YourFitnessFunctionRealValued(FitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.number_of_variables = number_of_variables
        self.value_to_reach = value_to_reach
        self.c_inst_realvalued = new yourFitnessFunctionRealValued(number_of_variables,value_to_reach)
    
    def __dealloc__(self):
        del self.c_inst_realvalued

cdef class GBOFitnessFunction(FitnessFunction):
    cpdef int number_of_subfunctions( self ) except -1:
        return -1
    
    cpdef vector[int] inputs_to_subfunction( self, int subfunction_index ) except *:
        return []
    
    cpdef double subfunction( self, int subfunction_index, np.ndarray variables ) except? INFINITY:
        return INFINITY
    
    cpdef double objective_function( self, int objective_index, np.ndarray fitness_buffers ) except? INFINITY:
        return fitness_buffers[objective_index]
    
    cpdef double constraint_function( self, np.ndarray fitness_buffers ) except? INFINITY:
        return 0
    
    cpdef int number_of_fitness_buffers( self ) except -1:
        return 1

    cpdef int fitness_buffer_index_for_subfunction( self, int subfunction_index ) except -1: 
        return 0

    cpdef double similarity_measure( self, size_t var_a, size_t var_b ) except? INFINITY:
        return -1

cdef class GBOFitnessFunctionDiscrete(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 1e308,
        alphabet_size : int = 2,
        **kwargs
    ):
        self.number_of_variables = number_of_variables
        self.alphabet_size = alphabet_size
        self.value_to_reach = value_to_reach
        self.c_inst_discrete = new pyGBOFitnessFunction_t[char](number_of_variables,alphabet_size,value_to_reach,<PyObject*>self)
    
    def __dealloc__(self):
        del self.c_inst_discrete

cdef class GBOFitnessFunctionRealValued(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0,
        **kwargs
    ):
        self.number_of_variables = number_of_variables
        self.value_to_reach = value_to_reach
        self.c_inst_realvalued = new pyGBOFitnessFunction_t[double](number_of_variables,value_to_reach,<PyObject*>self)
    
    def __dealloc__(self):
        del self.c_inst_realvalued

cdef class BBOFitnessFunction(FitnessFunction):
    cpdef double objective_function( self, int objective_index, np.ndarray variables ) except? INFINITY:
        return INFINITY
    
    cpdef double constraint_function( self, np.ndarray variables ) except? INFINITY:
        return 0

cdef class BBOFitnessFunctionDiscrete(BBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 1e308,
        alphabet_size : int = 2,
        **kwargs
    ):
        self.number_of_variables = number_of_variables
        self.alphabet_size = alphabet_size
        self.value_to_reach = value_to_reach
        self.c_inst_discrete = new pyBBOFitnessFunction_t[char](number_of_variables,alphabet_size,value_to_reach,<PyObject*>self)

    def __dealloc__(self):
        del self.c_inst_discrete

cdef class BBOFitnessFunctionRealValued(BBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0,
        **kwargs
    ):
        self.number_of_variables = number_of_variables
        self.value_to_reach = value_to_reach
        self.c_inst_realvalued = new pyBBOFitnessFunction_t[double](number_of_variables,value_to_reach,<PyObject*>self)
    
    def __dealloc__(self):
        del self.c_inst_realvalued


cdef class SphereFunction(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.c_inst_realvalued = new sphereFunction_t(number_of_variables,value_to_reach)
        self.number_of_variables = self.c_inst_realvalued.getNumberOfVariables()
        self.value_to_reach = self.c_inst_realvalued.getVTR()

cdef class SphereFunctionBBO(BBOFitnessFunction):
    def __cinit__(self,
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.c_inst_realvalued = new sphereFunctionBBO_t(number_of_variables,value_to_reach)
        self.number_of_variables = self.c_inst_realvalued.getNumberOfVariables()
        self.value_to_reach = self.c_inst_realvalued.getVTR()

cdef class RosenbrockFunction(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.c_inst_realvalued = new rosenbrockFunction_t(number_of_variables,value_to_reach)
        self.number_of_variables = self.c_inst_realvalued.getNumberOfVariables()
        self.value_to_reach = self.c_inst_realvalued.getVTR()

cdef class RosenbrockFunctionBBO(BBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.c_inst_realvalued = new rosenbrockFunctionBBO_t(number_of_variables,value_to_reach)
        self.number_of_variables = self.c_inst_realvalued.getNumberOfVariables()
        self.value_to_reach = self.c_inst_realvalued.getVTR()

cdef class SOREBChainStrong(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.c_inst_realvalued = new SOREBChainStrong_t(number_of_variables,value_to_reach)
        self.number_of_variables = self.c_inst_realvalued.getNumberOfVariables()
        self.value_to_reach = self.c_inst_realvalued.getVTR()

cdef class SOREBChainStrongBBO(BBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.c_inst_realvalued = new SOREBChainStrongBBO_t(number_of_variables,value_to_reach)
        self.number_of_variables = self.c_inst_realvalued.getNumberOfVariables()
        self.value_to_reach = self.c_inst_realvalued.getVTR()

cdef class CirclesInASquareBBO(BBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        value_to_reach : float = 0.0
    ):
        self.c_inst_realvalued = new circlesInASquareBBO_t(number_of_variables,value_to_reach)
        self.number_of_variables = self.c_inst_realvalued.getNumberOfVariables()
        self.value_to_reach = self.c_inst_realvalued.getVTR()


cdef class OneMaxFunction(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int
    ):
        self.c_inst_discrete = new oneMax_t(number_of_variables)
        self.number_of_variables = self.c_inst_discrete.getNumberOfVariables()
        self.value_to_reach = self.c_inst_discrete.getVTR()

cdef class DeceptiveTrapFunction(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        trap_size : int
    ):
        self.c_inst_discrete = new deceptiveTrap_t(number_of_variables,trap_size)
        self.number_of_variables = self.c_inst_discrete.getNumberOfVariables()
        self.value_to_reach = self.c_inst_discrete.getVTR()

cdef class DeceptiveTrapFunctionBBO(BBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        trap_size : int
    ):
        self.c_inst_discrete = new deceptiveTrapBBO_t(number_of_variables,trap_size)
        self.number_of_variables = self.c_inst_discrete.getNumberOfVariables()
        self.value_to_reach = self.c_inst_discrete.getVTR()

cdef class MaxCut(GBOFitnessFunction):
    def __cinit__(self,
        input_file : str,
        vtr_file : str = ""
    ):
        self.c_inst_discrete = new maxCut_t(str.encode(input_file),str.encode(vtr_file))
        self.number_of_variables = self.c_inst_discrete.getNumberOfVariables()
        self.value_to_reach = self.c_inst_discrete.getVTR()

cdef class MaxCutBBO(BBOFitnessFunction):
    def __cinit__(self,
        input_file : str,
        vtr_file : str = ""
    ):
        self.c_inst_discrete = new maxCutBBO_t(str.encode(input_file),str.encode(vtr_file))
        self.number_of_variables = self.c_inst_discrete.getNumberOfVariables()
        self.value_to_reach = self.c_inst_discrete.getVTR()

cdef class NKLandscapes(GBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        K : int = 5,
        seed : int = 0
    ):
        self.c_inst_discrete = new NKlandscapes_t(number_of_variables,K,seed)
        self.number_of_variables = self.c_inst_discrete.getNumberOfVariables()
        self.value_to_reach = self.c_inst_discrete.getVTR()

cdef class NKLandscapesBBO(BBOFitnessFunction):
    def __cinit__(self, 
        number_of_variables : int,
        K : int = 5,
        seed : int = 0
    ):
        self.c_inst_discrete = new NKlandscapesBBO_t(number_of_variables,K,seed)
        self.number_of_variables = self.c_inst_discrete.getNumberOfVariables()
        self.value_to_reach = self.c_inst_discrete.getVTR()



import gomea
import numpy as np
import cython
from functools import cache

# Custom fitness function resembling the concatenated deceptive trap function of size k
class MaxCut(gomea.fitness.GBOFitnessFunctionDiscrete):
    edges : np.ndarray
    def __new__(self, input_file, vtr_file ):
        number_of_variables : cython.int = 0
        value_to_reach : cython.double = 1e308
        self.edges : np.ndarray = []
        with open(input_file,"r") as f:
            lines = [line.split() for line in f.readlines()]
            number_of_variables = int(lines[0][0])
            number_of_edges = int(lines[0][1])
            for i in range(number_of_edges):
                v1 = int(lines[i+1][0])-1
                v2 = int(lines[i+1][1])-1
                w = float(lines[i+1][2])
                self.edges.append((v1,v2,w))
        with open(vtr_file,"r") as f:
            lines = f.readlines()
            value_to_reach = float(lines[0].strip())
        return super().__new__(self,number_of_variables,value_to_reach)

    @cache
    def number_of_subfunctions( self ) -> cython.int:
        return len(self.edges)
    
    @cache
    def inputs_to_subfunction( self, subfunction_index ) -> np.ndarray:
        return list(self.edges[subfunction_index])

    def subfunction(self, subfunction_index, variables) -> cython.double:
        v1 : cython.int = self.edges[subfunction_index][0]
        v2 : cython.int = self.edges[subfunction_index][1]
        w : cython.double = self.edges[subfunction_index][2]
        if variables[v1] == variables[v2]:
            return 0.0
        else:
            return w


import gomea
import numpy as np

# Custom GBO fitness function resembling the Rosenbrock function
class CustomRosenbrockFunction(gomea.fitness.GBOFitnessFunctionRealValued):
    def number_of_subfunctions( self ):
        return self.number_of_variables-1
    
    def inputs_to_subfunction( self, subfunction_index ):
        return [subfunction_index, subfunction_index+1]

    def subfunction(self, subfunction_index, variables):
        x = variables[subfunction_index]
        y = variables[subfunction_index+1]
        return 100*(y-x*x)*(y-x*x) + (1.0-x)*(1.0-x)

dim = 10
frv = CustomRosenbrockFunction(dim,value_to_reach=1e-6)
lm = gomea.linkage.Univariate()
rvgom = gomea.RealValuedGOMEA(fitness=frv, linkage_model=lm, lower_init_range=-115, upper_init_range=-100, max_number_of_populations=1, base_population_size=100, max_number_of_evaluations=1000000)
result = rvgom.run()
result.printAllStatistics()



from cpython cimport PyObject
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.string cimport string
from libcpp.map cimport map
from libc.math cimport INFINITY
import numpy as np
cimport numpy as np

cdef extern from "gomea/src/fitness/fitness.hpp" namespace "gomea::fitness":
    cdef cppclass fitness_t[T]:
        fitness_t() except +
        int getNumberOfVariables() except +
        double getVTR() except +

cdef extern from "gomea/src/fitness/gbo_fitness.hpp" namespace "gomea::fitness":
    cdef cppclass GBOFitnessFunction_t[T](fitness_t[T]):
        GBOFitnessFunction_t() except +

cdef extern from "gomea/src/fitness/bbo_fitness.hpp" namespace "gomea::fitness":
    cdef cppclass BBOFitnessFunction_t[T](fitness_t[T]):
        BBOFitnessFunction_t() except +

cdef extern from "gomea/src/fitness/py_gbo_fitness.hpp" namespace "gomea::fitness":
    cdef cppclass pyGBOFitnessFunction_t[T](GBOFitnessFunction_t[T]):
        pyGBOFitnessFunction_t() except +
        pyGBOFitnessFunction_t(int,PyObject*) except +
        pyGBOFitnessFunction_t(int,double,PyObject*) except +
        pyGBOFitnessFunction_t(int,int,double,PyObject*) except +

cdef extern from "gomea/src/fitness/py_bbo_fitness.hpp" namespace "gomea::fitness":
    cdef cppclass pyBBOFitnessFunction_t[T](fitness_t[T]):
        pyBBOFitnessFunction_t() except +
        pyBBOFitnessFunction_t(int,PyObject*) except +
        pyBBOFitnessFunction_t(int,double,PyObject*) except +
        pyBBOFitnessFunction_t(int,int,double,PyObject*) except +

cdef extern from "gomea/src/fitness/your_fitness_discrete.hpp" namespace "gomea::fitness":
    cdef cppclass yourFitnessFunctionDiscrete(GBOFitnessFunction_t[char]):
        yourFitnessFunctionDiscrete() except +
        yourFitnessFunctionDiscrete(int,int,double) except +

cdef extern from "gomea/src/fitness/your_fitness_realvalued.hpp" namespace "gomea::fitness":
    cdef cppclass yourFitnessFunctionRealValued(GBOFitnessFunction_t[double]):
        yourFitnessFunctionRealValued() except +
        yourFitnessFunctionRealValued(int,double) except +

cdef extern from "gomea/src/fitness/benchmarks-rv.hpp" namespace "gomea::fitness":
    cdef cppclass sphereFunction_t(GBOFitnessFunction_t[double]):
        sphereFunction_t() except +
        sphereFunction_t(int,double) except +

    cdef cppclass sphereFunctionBBO_t(BBOFitnessFunction_t[double]):
        sphereFunctionBBO_t() except +
        sphereFunctionBBO_t(int,double) except +

    cdef cppclass rosenbrockFunction_t(GBOFitnessFunction_t[double]):
        rosenbrockFunction_t() except +
        rosenbrockFunction_t(int,double) except +

    cdef cppclass SOREBChainStrong_t(GBOFitnessFunction_t[double]):
        SOREBChainStrong_t() except +
        SOREBChainStrong_t(int,double) except +
    
    cdef cppclass rosenbrockFunctionBBO_t(BBOFitnessFunction_t[double]):
        rosenbrockFunctionBBO_t() except +
        rosenbrockFunctionBBO_t(int,double) except +

    cdef cppclass SOREBChainStrongBBO_t(BBOFitnessFunction_t[double]):
        SOREBChainStrongBBO_t() except +
        SOREBChainStrongBBO_t(int,double) except +

    cdef cppclass circlesInASquareBBO_t(BBOFitnessFunction_t[double]):
        circlesInASquareBBO_t() except +
        circlesInASquareBBO_t(int,double) except +

cdef extern from "gomea/src/fitness/benchmarks-discrete.hpp" namespace "gomea::fitness":
    cdef cppclass oneMax_t(GBOFitnessFunction_t[char]):
        oneMax_t() except +
        oneMax_t(int) except +

    cdef cppclass deceptiveTrap_t(GBOFitnessFunction_t[char]):
        deceptiveTrap_t() except +
        deceptiveTrap_t(int,int) except +

    cdef cppclass deceptiveTrapBBO_t(BBOFitnessFunction_t[char]):
        deceptiveTrapBBO_t() except +
        deceptiveTrapBBO_t(int,int) except +

    cdef cppclass maxCut_t(GBOFitnessFunction_t[char]):
        maxCut_t() except +
        maxCut_t(string,string) except +

    cdef cppclass maxCutBBO_t(BBOFitnessFunction_t[char]):
        maxCutBBO_t() except +
        maxCutBBO_t(string,string) except +

    cdef cppclass NKlandscapes_t(GBOFitnessFunction_t[char]):
        NKlandscapes_t() except +
        NKlandscapes_t(int,int,long long) except +

    cdef cppclass NKlandscapesBBO_t(BBOFitnessFunction_t[char]):
        NKlandscapesBBO_t() except +
        NKlandscapesBBO_t(int,int,long long) except +

cdef class FitnessFunction:
    cdef public int number_of_variables
    cdef public double value_to_reach
    cdef map[pair[int,double],PyObject*] rotation_matrices
    cdef np.ndarray rotation_matrix

    cdef fitness_t[char] *c_inst_discrete
    cdef fitness_t[double] *c_inst_realvalued
    
    cpdef void initialize_rotation_matrix(self, int rotation_block_size, double rotation_angle)
    cpdef np.ndarray rotate_variables(self, np.ndarray variables, double rotation_angle)

cdef class GBOFitnessFunction(FitnessFunction):
    cpdef double subfunction( self, int subfunction_index, np.ndarray variables ) except? INFINITY
    cpdef double objective_function( self, int objective_index, np.ndarray fitness_buffers ) except? INFINITY
    cpdef double constraint_function( self, np.ndarray fitness_buffers ) except? INFINITY
    cpdef vector[int] inputs_to_subfunction( self, int ) except *
    cpdef int number_of_subfunctions( self ) except -1

    cpdef int number_of_fitness_buffers( self ) except -1
    cpdef int fitness_buffer_index_for_subfunction( self, int ) except -1
    
    cpdef double similarity_measure( self, size_t, size_t ) except? INFINITY

cdef class GBOFitnessFunctionDiscrete(GBOFitnessFunction):
    pass

cdef class GBOFitnessFunctionRealValued(GBOFitnessFunction):
    pass

cdef class BBOFitnessFunction(FitnessFunction):
    cpdef double objective_function( self, int objective_index, np.ndarray variables ) except? INFINITY
    cpdef double constraint_function( self, np.ndarray variables ) except? INFINITY

cdef class BBOFitnessFunctionDiscrete(BBOFitnessFunction):
    pass

cdef class BBOFitnessFunctionRealValued(BBOFitnessFunction):
    pass

